# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json


class RequestError(RuntimeError):
    """
    RequestError

    Parameters:
        message (str): The Request Error from Cosmos v5
        request (openc3.exceptions.CosmosJsonRpcRequest): CosmosJsonRpcRequest v5
    """

    def __init__(self, message: str, request):
        super().__init__(message, request)
        self.request = request


class ResponseError(RuntimeError):
    """
    ResponseError

    Parameters:
        request (openc3.exceptions.CosmosJsonRpcRequest): CosmosJsonRpcRequest v5
        response (openc3.exceptions.CosmosJsonRpcErrorResponse): CosmosJsonRpcErrorResponse v5
    """

    def __init__(self, request, response):
        super().__init__(request, response)
        self.request = request
        self.response = response


class JsonRpc(dict):
    def __init__(self):
        super().__init__()
        self["jsonrpc"] = "2.0"

    @property
    def id(self):
        return self.get("id")

    @property
    def json_rpc(self):
        return self.get("jsonrpc", "2.0")

    def to_json(self):
        return json.dumps(self)


class JsonRpcRequest(JsonRpc):
    """Represents a JSON Remote Procedure Call Request"""

    DANGEROUS_METHODS = ["__send__", "send", "instance_eval", "instance_exec"]

    def __init__(self, id: int, method_name: str, *args, **kwargs):
        """Constructor

        Arguments:
        id_ -- The identifier which will be matched to the response
        method_name -- The name of the method to call
        scope -- The scope
        args -- Array of strings which represent the parameters to send to the method
        kwargs -- Dict of Key, Value parameters to send to the method
        """
        super().__init__()
        self["method"] = method_name
        if args:
            self["params"] = args
        if kwargs:
            self["keyword_params"] = kwargs  # {**{"scope": scope}, **kwargs}
        self["id"] = int(id)

    @property
    def method(self):
        """Returns the method to call"""
        return self["method"]

    @property
    def params(self):
        """Returns the array of strings which represent the parameters to send to the method"""
        return self.get("params")

    @property
    def keyword_params(self):
        """Returns a dictionary of strings which represent the keyword parameters to send to the method"""
        return self.get("keyword_params")

    def to_hash(self):
        """Returns the request in a string"""
        return _convert_bytearray_to_string_raw(self)

    @classmethod
    def from_json(cls, request_data, request_headers):
        """Creates and returns a JsonRpcRequest object from a JSON encoded String.
        The version must be 2.0 and the JSON must include the method and id members.

        Parameters:
        request_data -- JSON encoded string representing the request
        request_headers -- Request header to include the auth token
        """
        msg = "invalid json-rpc 2.0 request"
        try:
            request = json.loads(request_data)
            if request_headers.get("HTTP_AUTHORIZATION"):
                request["keyword_params"]["token"] = request_headers["HTTP_AUTHORIZATION"]
            # Verify the jsonrpc version is correct and there is a method and id
            if request["jsonrpc"] != "2.0" or not request["method"] or not request["id"]:
                raise ValueError("message jsonrpc version: {}".format(request["jsonrpc"]))
            return cls.from_hash(request)
        except (ValueError, KeyError) as e:
            raise RequestError(msg, request_data) from e
        except Exception as e:
            raise RuntimeError(msg) from e

    @classmethod
    def from_hash(cls, request):
        """Creates a JsonRpcRequest object from a request dict

        Parameters:
        request -- dict containing the following keys: method, params, id, and keyword_params
        """
        return cls(
            request["id"],
            request["method"],
            *request.get("params", []),
            **request.get("keyword_params", {}),
        )


class JsonRpcResponse(JsonRpc):
    """Represents a JSON Remote Procedure Call Response"""

    def __init__(self, id):
        """Constructor

        Parameters:
        id_ -- The identifier which will be matched to the request
        """
        super().__init__()
        self["id"] = id

    @classmethod
    def from_json(cls, response_data):
        """Creates a JsonRpcResponse object from a JSON encoded String.

        The version must be 2.0 and the JSON must include the id members. It
        must also include either result for success or error for failure but
        never both.

        Parameters:
        response_data -- JSON encoded string representing the response
        """

        msg = f"invalid json-rpc 2.0 response{response_data}\n"
        if isinstance(response_data, str):
            try:
                return json.loads(response_data)  # .decode("latin-1"))
            except Exception as e:
                raise RuntimeError(msg, response_data) from e

    @classmethod
    def from_hash(cls, hash):
        # Verify the jsonrpc version is correct and there is an ID
        if hash.get("jsonrpc", None) != "2.0" or "id" not in hash:
            raise RuntimeError(f"invalid json-rpc 2.0 response:{hash}")

        if "result" in hash:
            if "error" in hash:
                raise RuntimeError(f"invalid json-rpc 2.0 response:{hash}")
            return JsonRpcSuccessResponse.from_hash(hash)
        elif "error" in hash:
            return JsonRpcErrorResponse.from_hash(hash)
        else:
            raise RuntimeError(f"invalid json-rpc 2.0 response:{hash}")


class JsonRpcSuccessResponse(JsonRpcResponse):
    """Represents a JSON Remote Procedure Call Success Response"""

    def __init__(self, id, result):
        """Constructor

        Parameters:
        id_ -- The identifier which will be matched to the request
        result -- The result of the request
        """
        super().__init__(id)
        result = convert_json_class(result)
        self["result"] = result

    @property
    def result(self):
        """Return the result of the method request"""
        return self["result"]

    @classmethod
    def from_hash(cls, hash):
        """Creates a JsonRpcSuccessResponse object from a Hash

        Parameters
        hash_ -- Hash containing the following keys: result and id
        """
        return cls(hash["id"], hash["result"])


class JsonRpcErrorResponse(JsonRpcResponse):
    """Represents a JSON Remote Procedure Call Error Response"""

    def __init__(self, id, error):
        """Constructor

        Parameters:
        id -- The identifier which will be matched to the request
        error -- The error object
        """
        super().__init__(id)
        self["error"] = JsonRpcError.from_hash(error)

    @property
    def error(self):
        """Returns the error object"""
        return self["error"]

    @classmethod
    def from_hash(cls, hash):
        """Creates a JsonRpcErrorResponse object from a Hash

        Parameters:
        hash -- Hash containing the following keys: error and id
        """
        return cls(hash["id"], hash["error"])


class JsonRpcError(dict):
    """Represents a JSON Remote Procedure Call Error"""

    def __init__(self, code, message, data=None):
        """Constructor

        Parameters:
        code -- The error type that occurred
        message -- A short description of the error
        data -- Additional information about the error
        """
        super().__init__()
        self["code"] = code
        self["message"] = message
        self["data"] = data

    @property
    def code(self):
        """Returns the error type that occurred"""
        return self.get("code")

    @property
    def message(self):
        """Returns a short description of the error"""
        return self.get("message")

    @property
    def data(self):
        """Returns additional information about the error"""
        return self.get("data")

    @classmethod
    def from_hash(cls, hash):
        """Creates a JsonRpcError object from a Hash

        Parameters:
        hash -- Hash containing the following keys: code, message, and optionally data
        """
        try:
            code = int(hash["code"])
            return cls(code, hash["message"], hash["data"])
        except ValueError as err:
            error = "Invalid JSON-RPC 2.0"
            raise RuntimeError("{} {}: {}".format(error, type(err), err)) from err


def convert_json_class(object_):
    if isinstance(object_, dict):
        try:
            json_class = object_["json_class"]
            raw = object_["raw"]
            if json_class == "Float":
                if raw == "Infinity":
                    return float("inf")
                elif raw == "-Infinity":
                    return float("-inf")
                elif raw == "NaN":
                    return float("nan")
                return raw
            return bytearray(raw)
        except Exception:
            for key, value in object_.items():
                object_[key] = convert_json_class(value)
            return object_
    elif isinstance(object_, (tuple, list)):
        object_ = list(object_)
        index = 0
        for value in object_:
            object_[index] = convert_json_class(value)
            index += 1
        return object_
    else:
        return object_


def _convert_bytearray_to_string_raw(object_):
    if isinstance(object_, (bytes, bytearray)):
        try:
            return object_.decode()
        except UnicodeDecodeError:
            return {"json_class": "String", "raw": [byte for byte in object_]}
    if isinstance(object_, dict):
        for key, value in object_.items():
            object_[key] = _convert_bytearray_to_string_raw(value)
        return object_
    if isinstance(object_, (tuple, list)):
        object_ = list(object_)
        index = 0
        for value in object_:
            object_[index] = _convert_bytearray_to_string_raw(value)
            index += 1
        return object_
    return object_



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from io import StringIO
from openc3.tools.table_manager.table_config import TableConfig
from openc3.utilities.string import simple_formatted

class TableManagerCore:
    """
    Provides the low level Table Manager methods which do not require a GUI.
    """

    class CoreError(Exception):
        """Generic error raised when a more specific error doesn't work"""
        pass

    class MismatchError(CoreError):
        """Raised when opening a file that is either larger or smaller than its definition"""
        pass

    @classmethod
    def binary(cls, binary, definition_filename, table_name):
        """
        Get the binary data for a specific table

        Args:
            binary: Binary data containing the table
            definition_filename: Table definition filename
            table_name: The specific table to return binary data for

        Returns:
            Table buffer
        """
        config = TableConfig.process_file(definition_filename)
        cls.load_binary(config, binary)
        return config.table(table_name).buffer

    @classmethod
    def definition(cls, definition_filename, table_name):
        """
        Get the definition for a specific table

        Args:
            definition_filename: Table definition filename
            table_name: The specific table to return the definition for

        Returns:
            Array containing [filename, contents]
        """
        config = TableConfig.process_file(definition_filename)
        return config.definition(table_name)  # This returns an array: [filename, contents]

    @classmethod
    def report(cls, binary, definition_filename, requested_table_name=None):
        """
        Create a text report of binary data for a given definition and optional table

        Args:
            binary: Binary data containing the table(s)
            definition_filename: Table definition filename
            requested_table_name: The specific table to return a report for (None = all tables)

        Returns:
            String report
        """
        report = StringIO()
        config = TableConfig.process_file(definition_filename)
        try:
            cls.load_binary(config, binary)
        except cls.CoreError as e:
            report.write(f"Error: {str(e)}\n")

        for table_name, table in config.tables.items():
            if requested_table_name and table_name != requested_table_name:
                continue
            items = table.sorted_items
            report.write(table.table_name + "\n")

            # Write the column headers
            if table.type == "ROW_COLUMN":
                columns = ['Item']

                # Remove the '0' from the 'itemname0'
                for x in range(table.num_columns):
                    columns.append(items[x].name[0:-1])
                report.write(", ".join(columns) + "\n")
            else:
                report.write("Label, Value\n")

            # Write the table item values
            for r in range(table.num_rows):
                if table.type == "ROW_COLUMN":
                    rowtext = f"{r + 1}"
                else:
                    rowtext = items[r].name

                report.write(f"{rowtext}, ")
                for c in range(table.num_columns):
                    if table.type == "ROW_COLUMN":
                        table_item = items[c + (r * table.num_columns)]
                    else:
                        table_item = items[r]
                    value = table.read(table_item.name, "FORMATTED")
                    if value.isascii():
                        report.write(f"{value}, ")
                    else:
                        report.write(f"{simple_formatted(value)}, ")
                report.write("\n")  # newline after each row
            report.write("\n")  # newline after each table

        return report.getvalue()

    @classmethod
    def generate(cls, definition_filename):
        """
        Generate binary data from a table definition file

        Args:
            definition_filename: Table definition filename

        Returns:
            Binary string representing the defined table(s)
        """
        config = TableConfig.process_file(definition_filename)
        binary = b''
        for _table_name, table in config.tables.items():
            table.restore_defaults()
            binary += table.buffer
        return binary

    @classmethod
    def save(cls, definition_filename, tables):
        """
        Update a table definition with table data and return binary data

        Args:
            definition_filename: Table definition filename
            tables: Hash of tables to save {'name' => name, 'rows' => array of rows}

        Returns:
            Binary data representing the defined table(s)
        """
        config = TableConfig.process_file(definition_filename)
        for table in tables:
            table_def = config.tables[table['name']]
            for row in table['rows']:
                for item in row:
                    # TODO: I don't know how the frontend could edit an item like this:
                    # item:{"name"=>"BINARY", "value"=>{"json_class"=>"String", "raw"=>[222, 173, 190, 239]} }
                    if isinstance(item['value'], dict):
                        continue
                    print("table_def:", table_def)
                    table_def.write(item['name'], item['value'])

        binary = b''
        for _table_name, table in config.tables.items():
            binary += table.buffer
        return binary

    @classmethod
    def build_json_hash(cls, binary, definition_filename):
        """
        Build JSON hash for the frontend to consume

        Args:
            binary: Binary data containing the table(s)
            definition_filename: Table definition filename

        Returns:
            Hash that can be converted to JSON
        """
        config = TableConfig.process_file(definition_filename)
        tables = []
        json_hash = {"tables": tables}
        try:
            cls.load_binary(config, binary)
        except cls.CoreError as e:
            json_hash['errors'] = str(e)

        for table_name, table in config.tables.items():
            tables.append({
                "name": table_name,
                "numRows": table.num_rows,
                "numColumns": table.num_columns,
                "headers": [],
                "rows": [],
            })

            col = 0
            row = 0
            for item in table.sorted_items:
                if item.hidden:
                    continue

                if table.num_columns == 1:
                    if row == 0:
                        tables[-1]["headers"] = ["INDEX", "NAME", "VALUE"]

                    tables[-1]["rows"].append([
                        {
                            "index": row + 1,
                            "name": item.name,
                            "value": table.read(item.name, "FORMATTED"),
                            "states": item.states,
                            "editable": item.editable,
                        },
                    ])
                else:
                    if row == 0 and col == 0:
                        tables[-1]["headers"].append("INDEX")

                    if row == 0:
                        tables[-1]["headers"].append(item.name[0:-1])

                    if col == 0:
                        # Each row is an array of items
                        tables[-1]["rows"].append([])

                    tables[-1]["rows"][row].append({
                        "index": row + 1,
                        "name": item.name,
                        "value": table.read(item.name, "FORMATTED"),
                        "states": item.states,
                        "editable": item.editable,
                    })

                col += 1
                if col == table.num_columns:
                    col = 0
                    row += 1

        return json_hash

    @classmethod
    def load_binary(cls, config, data):
        """
        Load binary data into the table

        Args:
            config: TableConfig instance
            data: Binary data containing the table(s)

        Raises:
            MismatchError if the data doesn't match the definition
        """
        binary_data_index = 0
        total_table_length = 0
        for _table_name, table in config.tables.items():
            total_table_length += table.length

        for _table_name, table in config.tables.items():
            if binary_data_index + table.length > len(data):
                table.buffer = data[binary_data_index:]
                raise cls.MismatchError(
                    f"Binary size of {len(data)} not large enough to fully represent table definition of length {total_table_length}. "
                    f"The remaining table definition (starting with byte {len(data) - binary_data_index} in {table.table_name}) will be filled with 0."
                )

            table.buffer = data[binary_data_index:binary_data_index + table.length]
            binary_data_index += table.length

        if binary_data_index < len(data):
            raise cls.MismatchError(
                f"Binary size of {len(data)} larger than table definition of length {total_table_length}. "
                f"Discarding the remaining {len(data) - binary_data_index} bytes."
            )


# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import threading
import queue
from time import sleep
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger

# See https://eclipse.dev/paho/files/paho.mqtt.python/html/client.html
import paho.mqtt.client as mqtt


class MqttStream:
    def __init__(self, hostname, port=1883, ssl=False, write_topic=None, read_topic=None, ack_timeout=5):
        super().__init__()

        self.hostname = hostname
        self.port = int(port)
        self.ssl = ConfigParser.handle_true_false(ssl)
        self.write_topic = ConfigParser.handle_none(write_topic)
        self.read_topic = ConfigParser.handle_none(read_topic)
        self.ack_timeout = float(ack_timeout)
        self.pkt_queue = queue.Queue()

        self.username = None
        self.password = None
        self.cert = None
        self.key = None
        self.ca_file = None
        self.keyfile_password = None

        # Mutex on write is needed to protect from commands coming in from more than one tool
        self.write_mutex = threading.RLock()

    # Connect the stream
    def connect(self):
        self.pkt_queue.empty()

        self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2)
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.user_data_set(self.pkt_queue)  # passed to on_message

        if self.ssl:
            self.client.tls_set()
        if self.username and self.password:
            self.client.username_pw_set(self.username, self.password)
        # You still need the ca_file if you're using your own cert and key
        if self.cert and self.key and self.ca_file:
            if self.keyfile_password:
                self.client.tls_set(
                    ca_certs=self.ca_file.name,
                    certfile=self.cert.name,
                    keyfile=self.key.name,
                    keyfile_password=self.keyfile_password,
                )
            else:
                self.client.tls_set(ca_certs=self.ca_file.name, certfile=self.cert.name, keyfile=self.key.name)
        elif self.ca_file:
            self.client.tls_set(ca_certs=self.ca_file.name)

        self.client.loop_start()
        # Connect doesn't fully establish the connection, it just sends the CONNECT packet
        # When the client loop receives an ONNACK packet from the broker in response to the CONNECT packet
        # it calls the on_connect callback and updates the client state to connected (is_connected() returns True)
        self.client.connect(self.hostname, self.port)
        i = 0
        while not self.client.is_connected() and i < (self.ack_timeout * 100):
            sleep(0.01)
            i += 1

    def on_connect(self, client, userdata, flags, reason_code, properties):
        if reason_code.is_failure:
            Logger.error(f"MQTT failed to connect: {reason_code}")
        else:
            # we should always subscribe from on_connect callback to be sure
            # our subscribed is persisted across reconnections.
            if self.read_topic:
                client.subscribe(self.read_topic)

    def connected(self):
        if self.client:
            return self.client.is_connected()
        else:
            return False

    def disconnect(self):
        if self.client:
            self.client.disconnect()
            self.client = None

    def on_message(self, client, userdata, message):
        # userdata is set via user_data_set
        userdata.put(message.payload)

    # @return [String] Returns a binary string of data from the read_topic
    def read(self):
        if not self.read_topic:
            raise RuntimeError("Attempt to read from write only stream")

        # No read mutex is needed because reads happen serially
        data = self.pkt_queue.get(block=True)
        if data is None or len(data) <= 0:
            if data is None:
                Logger.info("MqttStream: read returned None")
            if data is not None and len(data) <= 0:
                Logger.info("MqttStream: read returned 0 bytes")
            return None

        return data

    # @param data [String] A binary string of data to write to the write_topic
    def write(self, data):
        if not self.write_topic:
            raise RuntimeError("Attempt to write to read only stream")

        self.write_mutex.acquire()
        self.client.publish(self.write_topic, data)
        self.write_mutex.release()



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.config.config_parser import ConfigParser
from openc3.interfaces.protocols.terminated_protocol import TerminatedProtocol

# This file implements the SLIP protocol as documented in RFC 1055
# https://datatracker.ietf.org/doc/html/rfc1055

# SLIP is a framing protocol and is therefore expected to be used for packet deliniation


class SlipProtocol(TerminatedProtocol):
    # Note: Characters are expected to be given as integers
    # @param start_char   [Integer/nil] Character to place at the start of frames (Defaults to nil)
    # @param read_strip_characters [true/false] Strip off start_char and end_char from reads
    # @param read_enable_escaping [true/false] Whether to enable or disable character escaping on reads
    # @param write_enable_escaping [true/false] Whether to enable or disable character escaping on writes
    # @param end_char     [Integer] Character to place at the end of frames (Defaults to 0xC0)
    # @param esc_char     [Integer] Escape character (Defaults to 0xDB)
    # @param esc_end_char [Integer] Character to Escape End character (Defaults to 0xDC)
    # @param esc_esc_char [Integer] Character to Escape Escape character (Defaults to 0xDD)
    # @param allow_empty_data [true/false/nil] See Protocol#initialize
    def __init__(
        self,
        start_char=None,
        read_strip_characters=True,
        read_enable_escaping=True,
        write_enable_escaping=True,
        end_char="0xC0",
        esc_char="0xDB",
        esc_end_char="0xDC",
        esc_esc_char="0xDD",
        allow_empty_data=None,
    ):
        self.start_char = ConfigParser.handle_none(start_char)
        if self.start_char is not None:
            try:
                self.start_char = int(start_char, 0).to_bytes(1, byteorder="big")
            except ValueError:
                raise ValueError(f"invalid value {start_char} for start_char")
        self.end_char = int(end_char, 0).to_bytes(1, byteorder="big")
        self.esc_char = int(esc_char, 0).to_bytes(1, byteorder="big")
        self.esc_esc_char = int(esc_esc_char, 0).to_bytes(1, byteorder="big")
        self.esc_end_char = int(esc_end_char, 0).to_bytes(1, byteorder="big")
        self.replace_end = self.esc_char + self.esc_end_char
        self.replace_esc = self.esc_char + self.esc_esc_char
        self.read_strip_characters = ConfigParser.handle_true_false(read_strip_characters)
        if self.read_strip_characters is not True and self.read_strip_characters is not False:
            raise RuntimeError("read_strip_characters must be True or False")
        self.read_enable_escaping = ConfigParser.handle_true_false(read_enable_escaping)
        if self.read_enable_escaping is not True and self.read_enable_escaping is not False:
            raise RuntimeError("read_enable_escaping must be True or False")
        self.write_enable_escaping = ConfigParser.handle_true_false(write_enable_escaping)
        if self.write_enable_escaping is not True and self.write_enable_escaping is not False:
            raise RuntimeError("write_enable_escaping must be True or False")

        strip_read_termination = False
        discard_leading_bytes = 0
        sync_pattern = None
        if self.start_char:
            sync_pattern = hex(int(start_char, 0))
        fill_fields = False  # Handled in write_data below

        super().__init__(
            "",  # Write termination handled in write_data below
            hex(int(end_char, 0)),  # Expects Hex Character String
            strip_read_termination,
            discard_leading_bytes,
            sync_pattern,
            fill_fields,
            allow_empty_data,
        )

    def read_data(self, data, extra=None):
        data, extra = super().read_data(data, extra)
        if len(data) <= 0 or isinstance(data, str):
            return (data, extra)

        if self.read_strip_characters:
            if self.start_char:
                data = data[1:]
            data = data[0:-1]

        if self.read_enable_escaping:
            data = data.replace(self.replace_end, self.end_char).replace(self.replace_esc, self.esc_char)

        return (data, extra)

    def write_data(self, data, extra=None):
        # Intentionally not calling super()

        if self.write_enable_escaping:
            data = data.replace(self.esc_char, self.replace_esc).replace(self.end_char, self.replace_end)

        if self.start_char:
            data = self.start_char + data

        data += self.end_char

        return (data, extra)

    def reduce_to_single_packet(self):
        if len(self.data) <= 0:
            return ("STOP", self.extra)
        index = None
        if self.start_char is not None:
            try:
                index = self.data[1:].index(self.read_termination_characters)
                index = index + 1
            except ValueError:
                pass
        else:
            try:
                index = self.data.index(self.read_termination_characters)
            except ValueError:
                pass

        # Reduce to packet data and setup current_data for next packet
        if index is not None:
            if index > 0:
                packet_data = self.data[0 : (index + len(self.read_termination_characters))]
            else:  # self.data begins with the termination characters
                packet_data = self.data[0 : (len(self.read_termination_characters))]
            self.data = self.data[(index + len(self.read_termination_characters)) :]
            return (packet_data, self.extra)
        else:
            return ("STOP", self.extra)



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import sys
import math
import struct
from .accessor import Accessor


class BinaryAccessor(Accessor):
    # Constants for python struct packing directives
    # https://docs.python.org/3/library/struct.html
    STRUCT_INT_8 = "b"
    STRUCT_UINT_8 = "B"
    STRUCT_INT_16 = "h"
    STRUCT_UINT_16 = "H"
    STRUCT_INT_32 = "i"
    STRUCT_UINT_32 = "I"
    STRUCT_INT_64 = "q"
    STRUCT_UINT_64 = "Q"
    STRUCT_FLOAT_32 = "f"
    STRUCT_FLOAT_64 = "d"
    STRUCT_LITTLE_ENDIAN = "<"
    STRUCT_BIG_ENDIAN = ">"

    PACK_8_BIT_INT = "b"
    PACK_8_BIT_UINT = "B"
    PACK_BIG_ENDIAN_16_BIT_INT = ">h"
    PACK_LITTLE_ENDIAN_16_BIT_INT = "<h"
    PACK_BIG_ENDIAN_16_BIT_UINT = ">H"
    PACK_LITTLE_ENDIAN_16_BIT_UINT = "<H"
    PACK_BIG_ENDIAN_32_BIT_INT = ">i"
    PACK_LITTLE_ENDIAN_32_BIT_INT = "<i"
    PACK_NATIVE_32_BIT_UINT = "=I"
    PACK_BIG_ENDIAN_32_BIT_UINT = ">I"
    PACK_LITTLE_ENDIAN_32_BIT_UINT = "<I"
    PACK_BIG_ENDIAN_64_BIT_INT = ">q"
    PACK_LITTLE_ENDIAN_64_BIT_INT = "<q"
    PACK_BIG_ENDIAN_64_BIT_UINT = ">Q"
    PACK_LITTLE_ENDIAN_64_BIT_UINT = "<Q"
    PACK_BIG_ENDIAN_32_BIT_FLOAT = ">f"
    PACK_LITTLE_ENDIAN_32_BIT_FLOAT = "<f"
    PACK_BIG_ENDIAN_64_BIT_FLOAT = ">d"
    PACK_LITTLE_ENDIAN_64_BIT_FLOAT = "<d"
    MIN_INT8 = -128
    MAX_INT8 = 127
    MAX_UINT8 = 255
    MIN_INT16 = -32768
    MAX_INT16 = 32767
    MAX_UINT16 = 65535
    MIN_INT32 = -(2**31)
    MAX_INT32 = (2**31) - 1
    MAX_UINT32 = (2**32) - 1
    MIN_INT64 = -(2**63)
    MAX_INT64 = (2**63) - 1
    MAX_UINT64 = (2**64) - 1

    # Additional Constants
    ZERO_STRING = b"\00"

    # Valid data types
    DATA_TYPES = ["INT", "UINT", "FLOAT", "STRING", "BLOCK"]

    # Valid overflow types
    OVERFLOW_TYPES = ["TRUNCATE", "SATURATE", "ERROR", "ERROR_ALLOW_HEX"]

    # Determines the endianness of the host running this code
    #
    # This method is protected to force the use of the constant
    # HOST_ENDIANNESS rather than this method
    #
    # @return [Symbol] 'BIG_ENDIAN' or 'LITTLE_ENDIAN'
    @classmethod
    def get_host_endianness(cls):
        value = 0x01020304
        packed = struct.pack(BinaryAccessor.PACK_NATIVE_32_BIT_UINT, value)
        unpacked = struct.unpack(BinaryAccessor.PACK_LITTLE_ENDIAN_32_BIT_UINT, packed)
        if unpacked[0] == value:
            return "LITTLE_ENDIAN"
        else:
            return "BIG_ENDIAN"

    @classmethod
    def raise_buffer_error(cls, read_write, buffer, data_type, given_bit_offset, given_bit_size):
        raise ValueError(
            f"{len(buffer)} byte buffer insufficient to {read_write} {data_type} at bit_offset {given_bit_offset} with bit_size {given_bit_size}"
        )

    # Valid endianness
    ENDIANNESS = ["BIG_ENDIAN", "LITTLE_ENDIAN"]

    def handle_read_variable_bit_size(self, item, _buffer):
        length_value = self.packet.read(item.variable_bit_size["length_item_name"], "CONVERTED")
        if item.array_size is not None:
            item.array_size = (length_value * item.variable_bit_size["length_bits_per_count"]) + item.variable_bit_size[
                "length_value_bit_offset"
            ]
        else:
            if item.data_type == "INT" or item.data_type == "UINT":
                # QUIC encoding is currently assumed for individual variable sized integers
                # see https://datatracker.ietf.org/doc/html/rfc9000#name-variable-length-integer-enc
                match length_value:
                    case 0:
                        item.bit_size = 6
                    case 1:
                        item.bit_size = 14
                    case 2:
                        item.bit_size = 30
                    case _:
                        item.bit_size = 62
            else:
                item.bit_size = (
                    length_value * item.variable_bit_size["length_bits_per_count"]
                ) + item.variable_bit_size["length_value_bit_offset"]

    def read_item(self, item, buffer):
        if item.data_type == "DERIVED":
            return None
        if item.variable_bit_size:
            self.handle_read_variable_bit_size(item, buffer)
        return BinaryAccessor.class_read_item(item, buffer)

    # Note: do not use directly - use instance read_item
    @classmethod
    def class_read_item(cls, item, buffer):
        if item.data_type == "DERIVED":
            return None
        if item.array_size is not None:
            return cls.read_array(
                item.bit_offset,
                item.bit_size,
                item.data_type,
                item.array_size,
                buffer,
                item.endianness,
            )
        else:
            return cls.read(item.bit_offset, item.bit_size, item.data_type, buffer, item.endianness)

    def handle_write_variable_bit_size(self, item, value, buffer):
        adjustment = 0
        # Update length field to new size
        if (item.data_type == "INT" or item.data_type == "UINT") and item.original_array_size is None:
            adjustment = self._write_variable_int(item, value, buffer)
        # Probably not possible to get this condition because we don't allow 0 sized floats
        # but check for it just to cover all the possible data_types
        elif item.data_type == "FLOAT":
            raise ValueError("Variable bit size not currently supported for FLOAT data type")
        else:
            # STRING, BLOCK, or array types
            adjustment = self._write_variable_other(item, value, buffer)

        # Recalculate bit offsets after this item
        if adjustment != 0 and item.bit_offset >= 0:
            for sitem in self.packet.sorted_items:
                if sitem.data_type == "DERIVED" or sitem.bit_offset < item.bit_offset:
                    # Skip items before this item and derived items and items with negative bit offsets
                    continue
                if sitem != item:
                    sitem.bit_offset += adjustment

    def _write_variable_int(self, item, value, buffer):
        # QUIC encoding is currently assumed for individual variable sized integers
        # see https://datatracker.ietf.org/doc/html/rfc9000#name-variable-length-integer-enc

        # Calculate current bit size so we can preserve bytes after the item
        length_item_value = self.packet.read(item.variable_bit_size["length_item_name"], "CONVERTED")
        match length_item_value:
            case 0:
                current_bit_size = 6
            case 1:
                current_bit_size = 14
            case 2:
                current_bit_size = 30
            case 3:
                current_bit_size = 62
            case _:
                raise ValueError(
                    f"Value {item.variable_bit_size['length_item_name']} has unknown QUIC bit size encoding: {length_item_value}"
                )

        if item.data_type == "UINT":
            if value <= 63:
                # Length = 0, value up to 6-bits
                new_bit_size = 6
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 0)
            elif value <= 16383:
                # Length = 1, value up to 14-bits
                new_bit_size = 14
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 1)
            elif value <= 1073741823:
                # Length = 2, value up to 30-bits
                new_bit_size = 30
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 2)
            else:
                # Length = 3, value up to 62-bits
                new_bit_size = 62
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 3)
        else:
            if value <= 31 and value >= -32:
                # Length = 0, value up to 6-bits
                new_bit_size = 6
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 0)
            elif value <= 8191 and value >= -8192:
                # Length = 1, value up to 14-bits
                new_bit_size = 14
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 1)
            elif value <= 536870911 and value >= -536870912:
                # Length = 2, value up to 30-bits
                new_bit_size = 30
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 2)
            else:
                # Length = 3, value up to 62-bits
                new_bit_size = 62
                item.bit_size = new_bit_size
                self.packet.write(item.variable_bit_size["length_item_name"], 3)

        # Later items need their bit_offset adjusted by the change in this item
        adjustment = new_bit_size - current_bit_size
        adjust_bytes = int(adjustment / 8)
        item_offset = int(item.bit_offset / 8)
        if adjust_bytes > 0:
            original_length = len(buffer)
            # Add extra bytes because we're adjusting larger
            buffer += BinaryAccessor.ZERO_STRING * adjust_bytes
            # We added bytes to the end so now we have to shift the buffer over
            # We copy from the original offset with the original length
            # to the new shifted offset all the way to the end of the buffer
            buffer[(item_offset + adjust_bytes) :] = buffer[item_offset:original_length]
        elif adjust_bytes < 0:
            # Remove extra bytes because we're adjusting smaller
            del buffer[item_offset + 1 : item_offset + 1 - adjust_bytes]

        return adjustment

    def _write_variable_other(self, item, value, buffer):
        # Calculate current bit size so we can preserve bytes after the item
        length_item_value = self.packet.read(item.variable_bit_size["length_item_name"], "CONVERTED")
        current_bit_size = (
            length_item_value * item.variable_bit_size["length_bits_per_count"]
        ) + item.variable_bit_size["length_value_bit_offset"]

        # Calculate bits after this item
        bits_with_item = item.bit_offset + current_bit_size
        bits_after_item = (len(buffer) * 8) - bits_with_item
        if item.original_array_size is not None:
            item.array_size = -bits_after_item
        else:
            item.bit_size = -bits_after_item

        new_bit_size = len(value) * 8
        length_value = (new_bit_size - item.variable_bit_size["length_value_bit_offset"]) / item.variable_bit_size[
            "length_bits_per_count"
        ]
        self.packet.write(item.variable_bit_size["length_item_name"], length_value)

        # Later items need their bit_offset adjusted by the change in this item
        # so return the adjustment value
        return new_bit_size - current_bit_size

    def write_item(self, item, value, buffer):
        if item.data_type == "DERIVED":
            return None
        if item.variable_bit_size:
            self.handle_write_variable_bit_size(item, value, buffer)
        BinaryAccessor.class_write_item(item, value, buffer)

    # Note: do not use directly - use instance write_item
    @classmethod
    def class_write_item(cls, item, value, buffer):
        if item.data_type == "DERIVED":
            return None
        if item.array_size is not None:
            return cls.write_array(
                value,
                item.bit_offset,
                item.bit_size,
                item.data_type,
                item.array_size,
                buffer,
                item.endianness,
                item.overflow,
            )
        else:
            return cls.write(
                value,
                item.bit_offset,
                item.bit_size,
                item.data_type,
                buffer,
                item.endianness,
                item.overflow,
            )

    # Reads binary data of any data type from a buffer
    #
    # @param bit_offset [Integer] Bit offset to the start of the item. A
    #   negative number means to offset from the  of the buffer.
    # @param bit_size [Integer] Size of the item in bits
    # @param data_type [Symbol] {DATA_TYPES}
    # @param buffer [String] Binary string buffer to read from
    # @param endianness [Symbol] {ENDIANNESS}
    # @return [Integer] value read from the buffer
    @classmethod
    def read(cls, bit_offset, bit_size, data_type, buffer, endianness):
        given_bit_offset = bit_offset
        given_bit_size = bit_size

        bit_offset = cls.check_bit_offset_and_size("read", given_bit_offset, given_bit_size, data_type, buffer)

        # If passed a negative bit size with strings or blocks
        # recalculate based on the buffer length
        if (bit_size <= 0) and ((data_type == "STRING") or (data_type == "BLOCK")):
            bit_size = (len(buffer) * 8) - bit_offset + bit_size
            if bit_size == 0:
                return ""
            elif bit_size < 0:
                cls.raise_buffer_error("read", buffer, data_type, given_bit_offset, given_bit_size)

        result, lower_bound, upper_bound = cls.check_bounds_and_buffer_size(
            bit_offset, bit_size, len(buffer), endianness, data_type
        )
        if not result:
            cls.raise_buffer_error("read", buffer, data_type, given_bit_offset, given_bit_size)

        if data_type in ["STRING", "BLOCK"]:
            #######################################
            # Handle 'STRING' and 'BLOCK' data types
            #######################################

            if cls.byte_aligned(bit_offset):
                if data_type == "STRING":
                    try:
                        buffer = buffer[lower_bound : (upper_bound + 1)]
                        try:
                            return buffer[: buffer.index(BinaryAccessor.ZERO_STRING)].decode(encoding="utf-8")
                        except ValueError:
                            return buffer.decode(encoding="utf-8")
                    # If this 'STRING' contains binary buffer.decode will fail
                    # Instead of blowing up return the original buffer
                    except UnicodeDecodeError:
                        return buffer
                else:  # BLOCK
                    return buffer[lower_bound : upper_bound + 1]

            else:
                raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

        elif data_type in ["INT", "UINT"]:
            ###################################
            # Handle 'INT' and 'UINT' data types
            ###################################

            if cls.byte_aligned(bit_offset) and cls.even_bit_size(bit_size):
                # if data_type == "INT":
                ###########################################################
                # Handle byte-aligned 8, 16, 32, and 64 bit 'INT'
                ###########################################################
                const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                format = "%s%s" % (endian, const)
                return struct.unpack(
                    format,
                    buffer[lower_bound : upper_bound + 1],
                )[0]

            else:
                ##########################
                # Handle 'INT' and 'UINT' Bitfields
                ##########################

                # Extract Data for Bitfield
                if endianness == "LITTLE_ENDIAN":
                    # Bitoffset always refers to the most significant bit of a bitfield
                    num_bytes = math.floor((((bit_offset % 8) + bit_size - 1) / 8) + 1)
                    upper_bound = math.floor(bit_offset / 8)
                    lower_bound = upper_bound - num_bytes + 1

                    if lower_bound < 0:
                        raise ValueError(
                            f"LITTLE_ENDIAN bitfield with bit_offset {given_bit_offset} and bit_size {given_bit_size} is invalid"
                        )
                    temp_lower = lower_bound - 1 if lower_bound > 0 else lower_bound
                    temp_data = buffer[upper_bound : temp_lower or None : -1]
                else:
                    temp_data = buffer[lower_bound : upper_bound + 1]

                # Determine temp upper bound
                temp_upper = upper_bound - lower_bound

                # Handle Bitfield
                start_bits = bit_offset % 8
                start_mask = ~(0xFF << (8 - start_bits))
                total_bits = (temp_upper + 1) * 8
                right_shift = total_bits - start_bits - bit_size

                # Mask off unwanted bits at beginning
                temp = temp_data[0] & start_mask

                if upper_bound > lower_bound:
                    # Combine bytes into a FixNum
                    for temp_value in temp_data[1 : (temp_upper + 1)]:
                        temp = temp << 8
                        temp = temp + temp_value

                # Shift off unwanted bits at end
                temp = temp >> right_shift
                if data_type == "INT":
                    # Convert to negative if necessary
                    if (bit_size > 1) and (temp & (1 << (bit_size - 1))):
                        temp = -((1 << bit_size) - temp)
                return temp

        elif data_type == "FLOAT":
            ##########################
            # Handle 'FLOAT' data type
            ##########################

            if cls.byte_aligned(bit_offset):
                if bit_size in [32, 64]:
                    const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                    endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                    format = "%s%s" % (endian, const)
                    return struct.unpack(
                        format,
                        buffer[lower_bound : upper_bound + 1],
                    )[0]
                else:
                    raise ValueError(f"bit_size is {given_bit_size} but must be 32 or 64 for data_type {data_type}")
            else:
                raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

        else:
            ############################
            # Handle Unknown data types
            ############################
            raise TypeError(f"data_type {data_type} is not recognized")

    # Writes binary data of any data type to a buffer
    #
    # @param value [Varies] Value to write into the buffer
    # @param bit_offset [Integer] Bit offset to the start of the item. A
    #   negative number means to offset from the  of the buffer.
    # @param bit_size [Integer] Size of the item in bits
    # @param data_type [Symbol] {DATA_TYPES}
    # @param buffer [String] Binary string buffer to write to
    # @param endianness [Symbol] {ENDIANNESS}
    # @param overflow [Symbol] {OVERFLOW_TYPES}
    # @return [Integer] value passed in as a parameter
    @classmethod
    def write(cls, value, bit_offset, bit_size, data_type, buffer, endianness, overflow):
        given_bit_offset = bit_offset
        given_bit_size = bit_size

        bit_offset = cls.check_bit_offset_and_size("write", given_bit_offset, given_bit_size, data_type, buffer)

        # If passed a negative bit size with strings or blocks
        # recalculate based on the value length in bytes
        if (bit_size <= 0) and ((data_type == "STRING") or (data_type == "BLOCK")):
            bit_size = len(value) * 8

        result, lower_bound, upper_bound = cls.check_bounds_and_buffer_size(
            bit_offset, bit_size, len(buffer), endianness, data_type
        )
        if not result and (given_bit_size > 0):
            cls.raise_buffer_error("write", buffer, data_type, given_bit_offset, given_bit_size)

        # Check overflow type
        if (
            (overflow != "TRUNCATE")
            and (overflow != "SATURATE")
            and (overflow != "ERROR")
            and (overflow != "ERROR_ALLOW_HEX")
        ):
            raise ValueError(f"unknown overflow type {overflow}")

        if (data_type == "STRING") or (data_type == "BLOCK"):
            #######################################
            # Handle 'STRING' and 'BLOCK' data types
            #######################################

            if cls.byte_aligned(bit_offset):
                if isinstance(value, str):
                    temp = value.encode(encoding="utf-8")
                else:
                    temp = value
                if given_bit_size <= 0:
                    end_bytes = -math.floor(given_bit_size / 8)
                    old_upper_bound = len(buffer) - 1 - end_bytes
                    # Lower bound + end_bytes can never be more than 1 byte outside of the given buffer
                    if (lower_bound + end_bytes) > len(buffer):
                        cls.raise_buffer_error("write", buffer, data_type, given_bit_offset, given_bit_size)

                    if old_upper_bound < lower_bound:
                        # String was completely empty
                        if end_bytes > 0:
                            # Preserve bytes at end of buffer
                            buffer += BinaryAccessor.ZERO_STRING * len(value)
                            lower_index = lower_bound + len(value)
                            buffer[lower_index : lower_index + end_bytes] = buffer[
                                lower_bound : lower_bound + end_bytes
                            ]

                    elif bit_size == 0:
                        # Remove entire string
                        buffer[lower_bound : old_upper_bound + 1] = bytearray(b"")
                    elif upper_bound < old_upper_bound:
                        # Remove extra bytes from old string
                        buffer[upper_bound + 1 : old_upper_bound + 1] = bytearray(b"")
                    elif (upper_bound > old_upper_bound) and (end_bytes > 0):
                        # Preserve bytes at end of buffer
                        diff = upper_bound - old_upper_bound
                        buffer += BinaryAccessor.ZERO_STRING * diff
                        buffer[upper_bound + 1 : upper_bound + 1 + end_bytes] = buffer[
                            old_upper_bound + 1 : old_upper_bound + 1 + end_bytes
                        ]

                else:  # given_bit_size > 0
                    byte_size = math.floor(bit_size / 8)
                    if len(value) < byte_size:
                        if isinstance(value, str):
                            ba = bytearray()
                            ba.extend(value.encode(encoding="utf-8"))
                            value = ba
                        # Pad the requested size with zeros
                        temp = value.ljust(byte_size, BinaryAccessor.ZERO_STRING)
                    elif len(value) > byte_size:
                        if overflow == "TRUNCATE":
                            # Resize the value to fit the field
                            temp = value[0:byte_size]
                        else:
                            raise ValueError(
                                f"value of {len(value)} bytes does not fit into {byte_size} bytes for data_type {data_type}"
                            )

                if bit_size != 0:
                    buffer[lower_bound : lower_bound + len(temp)] = temp
            else:
                raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

        elif (data_type == "INT") or (data_type == "UINT"):
            ###################################
            # Handle 'INT' data type
            ###################################
            value = int(value)
            min_value, max_value, hex_max_value = cls.get_check_overflow_ranges(bit_size, data_type)
            value = cls.check_overflow(
                value,
                min_value,
                max_value,
                hex_max_value,
                bit_size,
                data_type,
                overflow,
            )

            if cls.byte_aligned(bit_offset) and cls.even_bit_size(bit_size):
                ###########################################################
                # Handle byte-aligned 8, 16, 32, and 64 bit
                ###########################################################
                const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                format = "%s%s" % (endian, const)
                buffer[lower_bound : upper_bound + 1] = struct.pack(
                    format,
                    value,
                )

            else:
                ###########################################################
                # Handle bit fields
                ###########################################################

                # Extract Existing Data
                if endianness == "LITTLE_ENDIAN":
                    # Bitoffset always refers to the most significant bit of a bitfield
                    num_bytes = math.floor(((bit_offset % 8) + bit_size - 1) / 8) + 1
                    upper_bound = math.floor(bit_offset / 8)
                    lower_bound = upper_bound - num_bytes + 1
                    if lower_bound < 0:
                        raise ValueError(
                            f"LITTLE_ENDIAN bitfield with bit_offset {given_bit_offset} and bit_size {given_bit_size} is invalid"
                        )

                    temp_lower = lower_bound - 1 if lower_bound > 0 else lower_bound
                    temp_data = buffer[upper_bound : temp_lower or None : -1]
                else:
                    temp_data = buffer[lower_bound : upper_bound + 1]

                # Determine temp upper bound
                temp_upper = upper_bound - lower_bound

                # Determine Values needed to Handle Bitfield
                start_bits = bit_offset % 8
                start_mask = 0xFF << (8 - start_bits)
                total_bits = (temp_upper + 1) * 8
                end_bits = total_bits - start_bits - bit_size
                end_mask = ~(0xFF << end_bits)

                # Add in Start Bits
                temp = int.from_bytes(temp_data[0:1], sys.byteorder) & start_mask

                # Adjust value to correct number of bits
                temp_mask = (2**bit_size) - 1
                temp_value = value & temp_mask

                shift = bit_size - (8 - start_bits)
                # Add in New Data
                if shift < 0:
                    temp = (temp >> -shift) + temp_value
                else:
                    temp = (temp << shift) + temp_value

                # Add in Remainder of Existing Data
                temp = (temp << end_bits) + (temp_data[temp_upper] & end_mask)

                # Extract into an array of bytes
                temp_array = bytearray()
                for _ in range(0, temp_upper + 1):
                    temp_array.insert(0, (temp & 0xFF))
                    temp = temp >> 8

                # Store into data
                if endianness == "LITTLE_ENDIAN":
                    temp_lower = lower_bound - 1 if lower_bound > 0 else lower_bound
                    temp_array.reverse()
                    # buffer[upper_bound : temp_lower or None : -1] = struct.pack(
                    buffer[lower_bound : upper_bound + 1] = struct.pack(
                        f"{len(temp_array)}{BinaryAccessor.PACK_8_BIT_UINT}",
                        *temp_array,
                    )
                else:
                    buffer[lower_bound : upper_bound + 1] = struct.pack(
                        f"{len(temp_array)}{BinaryAccessor.PACK_8_BIT_UINT}",
                        *temp_array,
                    )

        elif data_type == "FLOAT":
            ##########################
            # Handle 'FLOAT' data type
            ##########################
            value = float(value)

            if cls.byte_aligned(bit_offset):
                if bit_size in [32, 64]:
                    const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                    endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                    format = "%s%s" % (endian, const)
                    buffer[lower_bound : upper_bound + 1] = struct.pack(
                        format,
                        value,
                    )
                else:
                    raise ValueError(f"bit_size is {given_bit_size} but must be 32 or 64 for data_type {data_type}")
            else:
                raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

        else:
            ############################
            # Handle Unknown data types
            ############################

            raise TypeError(f"data_type {data_type} is not recognized")

        return value

    # Check the bit size and bit offset for problems. Recalculate the bit offset
    # and return back through the passed in pointer.
    @classmethod
    def check_bit_offset_and_size(cls, read_or_write, given_bit_offset, given_bit_size, data_type, buffer):
        bit_offset = given_bit_offset

        if (given_bit_size <= 0) and (data_type != "STRING") and (data_type != "BLOCK"):
            raise ValueError(
                f"bit_size {given_bit_size} must be positive for data types other than 'STRING' and 'BLOCK'"
            )

        if (given_bit_size <= 0) and (given_bit_offset < 0):
            raise ValueError(
                f"negative or zero bit_sizes ({given_bit_size}) cannot be given with negative bit_offsets ({given_bit_offset})"
            )

        if given_bit_offset < 0:
            bit_offset = (len(buffer) * 8) + bit_offset
            if bit_offset < 0:
                cls.raise_buffer_error(read_or_write, buffer, data_type, given_bit_offset, given_bit_size)

        return bit_offset

    # Calculate the bounds of the string to access the item based on the bit_offset and bit_size.
    # Also determine if the buffer size is sufficient.
    @classmethod
    def check_bounds_and_buffer_size(cls, bit_offset, bit_size, buffer_length, endianness, data_type):
        result = True  # Assume ok

        # Define bounds of string to access this item
        lower_bound = math.floor(bit_offset / 8)
        upper_bound = math.floor((bit_offset + bit_size - 1) / 8)

        # Sanity check buffer size
        if upper_bound >= buffer_length:
            # If it's not the special match of little endian bit field then we fail and return false
            if not (
                (endianness == "LITTLE_ENDIAN")
                and ((data_type == "INT") or (data_type == "UINT"))
                and (
                    # Not byte aligned with an even bit size
                    not ((cls.byte_aligned(bit_offset)) and (cls.even_bit_size(bit_size)))
                )
                and (lower_bound < buffer_length)
            ):
                result = False

        return result, lower_bound, upper_bound

    @classmethod
    def get_check_overflow_ranges(cls, bit_size, data_type):
        min_value = 0  # Default for UINT cases

        match bit_size:
            case 8:
                hex_max_value = BinaryAccessor.MAX_UINT8
                if data_type == "INT":
                    min_value = BinaryAccessor.MIN_INT8
                    max_value = BinaryAccessor.MAX_INT8
                else:
                    max_value = BinaryAccessor.MAX_UINT8

            case 16:
                hex_max_value = BinaryAccessor.MAX_UINT16
                if data_type == "INT":
                    min_value = BinaryAccessor.MIN_INT16
                    max_value = BinaryAccessor.MAX_INT16
                else:
                    max_value = BinaryAccessor.MAX_UINT16

            case 32:
                hex_max_value = BinaryAccessor.MAX_UINT32
                if data_type == "INT":
                    min_value = BinaryAccessor.MIN_INT32
                    max_value = BinaryAccessor.MAX_INT32
                else:
                    max_value = BinaryAccessor.MAX_UINT32

            case 64:
                hex_max_value = BinaryAccessor.MAX_UINT64
                if data_type == "INT":
                    min_value = BinaryAccessor.MIN_INT64
                    max_value = BinaryAccessor.MAX_INT64
                else:
                    max_value = BinaryAccessor.MAX_UINT64

            case _:  # Bitfield
                if data_type == "INT":
                    # Note signed integers must allow up to the maximum unsigned value to support values given in hex
                    if bit_size > 1:
                        max_value = 2 ** (bit_size - 1)
                        # min_value = -(2 ** bit_size - 1)
                        min_value = -max_value
                        # max_value = (2 ** bit_size - 1) - 1
                        max_value -= 1
                        # hex_max_value = (2 ** bit_size) - 1
                        hex_max_value = (2**bit_size) - 1
                    else:  # 1-bit signed
                        min_value = -1
                        max_value = 1
                        hex_max_value = 1

                else:
                    max_value = (2**bit_size) - 1
                    hex_max_value = max_value

        return min_value, max_value, hex_max_value

    @classmethod
    def byte_aligned(cls, value):
        return (value % 8) == 0

    @classmethod
    def even_bit_size(cls, bit_size):
        return (bit_size == 8) or (bit_size == 16) or (bit_size == 32) or (bit_size == 64)

    # Reads an array of binary data of any data type from a buffer
    #
    # @param bit_offset [Integer] Bit offset to the start of the array. A
    #   negative number means to offset from the  of the buffer.
    # @param bit_size [Integer] Size of each item in the array in bits
    # @param data_type [Symbol] {DATA_TYPES}
    # @param array_size [Integer] Size in bits of the array. 0 or negative means
    #   fill the array with as many bit_size number of items that exist (negative
    #   means excluding the final X number of bits).
    # @param buffer [String] Binary string buffer to read from
    # @param endianness [Symbol] {ENDIANNESS}
    # @return [Array] Array created from reading the buffer
    @classmethod
    def read_array(cls, bit_offset, bit_size, data_type, array_size, buffer, endianness):
        if len(buffer) == 0:
            return []

        # Save given values of bit offset, bit size, and array_size
        given_bit_offset = bit_offset
        given_bit_size = bit_size
        given_array_size = array_size

        # Handle negative and zero bit sizes
        if bit_size <= 0:
            raise ValueError(f"bit_size {given_bit_size} must be positive for arrays")

        # Handle negative bit offsets
        if bit_offset < 0:
            bit_offset = (len(buffer) * 8) + bit_offset
            if bit_offset < 0:
                cls.raise_buffer_error("read", buffer, data_type, given_bit_offset, given_bit_size)

        # Handle negative and zero array sizes
        if array_size <= 0:
            if given_bit_offset < 0:
                raise ValueError(
                    f"negative or zero array_size ({given_array_size}) cannot be given with negative bit_offset ({given_bit_offset})"
                )
            else:
                array_size = (len(buffer) * 8) - bit_offset + array_size
                if array_size == 0:
                    return []
                elif array_size < 0:
                    cls.raise_buffer_error("read", buffer, data_type, given_bit_offset, given_bit_size)

        # Calculate number of items in the array
        # If there is a remainder then we have a problem
        if array_size % bit_size != 0:
            raise ValueError(f"array_size {given_array_size} not a multiple of bit_size {given_bit_size}")

        num_items = math.floor(array_size / bit_size)

        # Define bounds of string to access this item
        lower_bound = math.floor(bit_offset / 8)
        upper_bound = math.floor((bit_offset + array_size - 1) / 8)

        # Check for byte alignment
        byte_aligned = (bit_offset % 8) == 0

        match data_type:
            case "STRING" | "BLOCK":
                #######################################
                # Handle 'STRING' and 'BLOCK' data types
                #######################################

                if byte_aligned:
                    value = []
                    for _ in range(num_items):
                        value.append(cls.read(bit_offset, bit_size, data_type, buffer, endianness))
                        bit_offset += bit_size
                else:
                    raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

            case "INT" | "UINT":
                ###################################
                # Handle 'INT' and 'UINT' data types
                ###################################
                if byte_aligned and (bit_size == 8 or bit_size == 16 or bit_size == 32 or bit_size == 64):
                    ###########################################################
                    # Handle byte-aligned 8, 16, 32, and 64 bit 'INT' and 'UINT'
                    ###########################################################
                    const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                    endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                    format = "%s%d%s" % (endian, num_items, const)
                    return list(
                        struct.unpack(
                            format,
                            buffer[lower_bound : upper_bound + 1],
                        )
                    )
                else:
                    ##################################
                    # Handle 'INT' and 'UINT' Bitfields
                    ##################################
                    if endianness == "LITTLE_ENDIAN" and bit_size > 1:
                        raise ValueError(
                            "read_array does not support little endian bit fields with bit_size greater than 1-bit"
                        )

                    value = []
                    for _ in range(num_items):
                        value.append(cls.read(bit_offset, bit_size, data_type, buffer, endianness))
                        bit_offset += bit_size

            case "FLOAT":
                ##########################
                # Handle 'FLOAT' data type
                ##########################

                if byte_aligned:
                    if bit_size in [32, 64]:
                        const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                        endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                        format = "%s%d%s" % (endian, num_items, const)
                        return list(
                            struct.unpack(
                                format,
                                buffer[lower_bound : upper_bound + 1],
                            )
                        )
                    else:
                        raise ValueError(
                            f"bit_size is {given_bit_size} but must be 32 or 64 for data_type {data_type}"
                        )

                else:
                    raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

            case _:
                ############################
                # Handle Unknown data types
                ############################

                raise TypeError(f"data_type {data_type} is not recognized")

        return list(value)

    # Writes an array of binary data of any data type to a buffer
    #
    # @param values [Array] Values to write into the buffer
    # @param bit_offset [Integer] Bit offset to the start of the array. A
    #   negative number means to offset from the  of the buffer.
    # @param bit_size [Integer] Size of each item in the array in bits
    # @param data_type [Symbol] {DATA_TYPES}
    # @param array_size [Integer] Size in bits of the array as represented in the buffer.
    #   Size 0 means to fill the buffer with as many bit_size number of items that exist
    #   (negative means excluding the final X number of bits).
    # @param buffer [String] Binary string buffer to write to
    # @param endianness [Symbol] {ENDIANNESS}
    # @return [Array] values passed in as a parameter
    @classmethod
    def write_array(
        cls,
        values,
        bit_offset,
        bit_size,
        data_type,
        array_size,
        buffer,
        endianness,
        overflow,
    ):
        # Save given values of bit offset, bit size, and array_size
        given_bit_offset = bit_offset
        given_bit_size = bit_size
        given_array_size = array_size

        # Verify a list was given
        if not isinstance(values, list):
            raise TypeError(f"values must be a list but is {values.__class__.__name__}")

        # Handle negative and zero bit sizes
        if bit_size <= 0:
            raise ValueError(f"bit_size {given_bit_size} must be positive for arrays")

        # Handle negative bit offsets
        if bit_offset < 0:
            bit_offset = (len(buffer) * 8) + bit_offset
            if bit_offset < 0:
                cls.raise_buffer_error("write", buffer, data_type, given_bit_offset, given_bit_size)

        # Handle negative and zero array sizes
        if array_size <= 0:
            if given_bit_offset < 0:
                raise ValueError(
                    f"negative or zero array_size ({given_array_size}) cannot be given with negative bit_offset ({given_bit_offset})"
                )
            else:
                end_bytes = -math.floor(given_array_size / 8)
                lower_bound = math.floor(bit_offset / 8)
                upper_bound = math.floor((bit_offset + (bit_size * len(values)) - 1) / 8)
                old_upper_bound = len(buffer) - 1 - end_bytes

                if upper_bound < old_upper_bound:
                    # Remove extra bytes from old buffer
                    buffer[(upper_bound + 1) : old_upper_bound + 1] = b""
                elif upper_bound > old_upper_bound:
                    # Grow buffer and preserve bytes at  of buffer if necessary
                    buffer_length = len(buffer)
                    diff = upper_bound - old_upper_bound
                    buffer += BinaryAccessor.ZERO_STRING * diff
                    if end_bytes > 0:
                        buffer[(upper_bound + 1) : len(buffer)] = buffer[(old_upper_bound + 1) : buffer_length]

                array_size = (len(buffer) * 8) - bit_offset + array_size

        # Get data bounds for this array
        lower_bound = math.floor(bit_offset / 8)
        upper_bound = math.floor((bit_offset + array_size - 1) / 8)

        # Check for byte alignment
        byte_aligned = (bit_offset % 8) == 0

        # Calculate the number of writes
        num_writes = math.floor(array_size / bit_size)
        # Check for a negative array_size and adjust the number of writes
        # to simply be the number of values in the passed in array
        if given_array_size <= 0:
            num_writes = len(values)

        # Ensure the buffer has enough room
        if bit_offset + num_writes * bit_size > len(buffer) * 8:
            cls.raise_buffer_error("write", buffer, data_type, given_bit_offset, given_bit_size)

        # Ensure the given_array_size is an even multiple of bit_size
        if array_size % bit_size != 0:
            raise ValueError(f"array_size {given_array_size} not a multiple of bit_size {given_bit_size}")
        if num_writes < len(values):
            raise ValueError(
                f"too many values {len(values)} for given array_size {given_array_size} and bit_size {given_bit_size}"
            )

        # Check overflow type
        if overflow not in BinaryAccessor.OVERFLOW_TYPES:
            raise TypeError(f"unknown overflow type {overflow}")

        # Expand the values by appending 0
        if len(values) < num_writes:
            for _ in range(0, (num_writes - len(values))):
                values.append(0)

        match data_type:
            case "STRING" | "BLOCK":
                #######################################
                # Handle 'STRING' and 'BLOCK' data types
                #######################################

                if byte_aligned:
                    for index in range(num_writes):
                        value = values[index]
                        if isinstance(value, int):
                            value = value.to_bytes(1, byteorder="big")
                        cls.write(
                            value,
                            bit_offset,
                            bit_size,
                            data_type,
                            buffer,
                            endianness,
                            overflow,
                        )
                        bit_offset += bit_size
                else:
                    raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

            case "INT" | "UINT":
                ###################################
                # Handle 'INT' and 'UINT' data types
                ###################################

                if byte_aligned and (bit_size == 8 or bit_size == 16 or bit_size == 32 or bit_size == 64):
                    ###########################################################
                    # Handle byte-aligned 8, 16, 32, and 64 bit 'INT' and 'UINT'
                    ###########################################################
                    max_hex = getattr(BinaryAccessor, f"MAX_UINT{bit_size}")
                    if data_type == "INT":
                        min_val = getattr(BinaryAccessor, f"MIN_INT{bit_size}")
                        max_val = getattr(BinaryAccessor, f"MAX_INT{bit_size}")
                    else:
                        min_val = 0
                        max_val = max_hex
                    values = cls.check_overflow_array(
                        values,
                        min_val,
                        max_val,
                        max_hex,
                        bit_size,
                        data_type,
                        overflow,
                    )
                    const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                    endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                    format = "%s%d%s" % (endian, num_writes, const)
                    buffer[lower_bound : upper_bound + 1] = struct.pack(
                        format,
                        *values,
                    )

                else:
                    ##################################
                    # Handle 'INT' and 'UINT' Bitfields
                    ##################################
                    if endianness == "LITTLE_ENDIAN" and bit_size > 1:
                        raise ValueError(
                            "write_array does not support little endian bit fields with bit_size greater than 1-bit"
                        )

                    for index in range(num_writes):
                        cls.write(
                            values[index],
                            bit_offset,
                            bit_size,
                            data_type,
                            buffer,
                            endianness,
                            overflow,
                        )
                        bit_offset += bit_size

            case "FLOAT":
                ##########################
                # Handle 'FLOAT' data type
                ##########################

                if byte_aligned:
                    if bit_size in [32, 64]:
                        const = getattr(BinaryAccessor, f"STRUCT_{data_type}_{bit_size}")
                        endian = getattr(BinaryAccessor, f"STRUCT_{endianness}")
                        format = "%s%d%s" % (endian, num_writes, const)
                        buffer[lower_bound : upper_bound + 1] = struct.pack(
                            format,
                            *values,
                        )
                    else:
                        raise ValueError(
                            f"bit_size is {given_bit_size} but must be 32 or 64 for data_type {data_type}"
                        )
                else:
                    raise ValueError(f"bit_offset {given_bit_offset} is not byte aligned for data_type {data_type}")

            case _:
                ############################
                # Handle Unknown data types
                ############################
                raise TypeError(f"data_type {data_type} is not recognized")

    #   # Adjusts the packed array to be the given number of bytes
    #   #
    #   # @param num_bytes [Integer] The desired number of bytes
    #   # @param packed [Array] The packed data buffer
    #   def self.adjust_packed_size(num_bytes, packed)
    #     difference = num_bytes - len(packed)
    #     if difference > 0
    #       packed << (ZERO_STRING * difference)
    #     elif difference < 0
    #       packed = packed[0..(len(packed) - 1 + difference)]

    #     packed

    #   # Byte swaps every X bytes of data in a buffer overwriting the buffer
    #   #
    #   # @param buffer [String] Buffer to modify
    #   # @param num_bytes_per_word [Integer] Number of bytes per word that will be swapped
    #   # @return [String] buffer passed in as a parameter
    #   def self.byte_swap_buffer!(buffer, num_bytes_per_word)
    #     num_swaps = len(buffer) / num_bytes_per_word
    #     index = 0
    #     num_swaps.times do
    #       range = index..(index + num_bytes_per_word - 1)
    #       buffer[range] = buffer[range].reverse
    #       index += num_bytes_per_word

    #     buffer

    #   # Byte swaps every X bytes of data in a buffer into a new buffer
    #   #
    #   # @param buffer [String] Buffer that will be copied then modified
    #   # @param num_bytes_per_word [Integer] Number of bytes per word that will be swapped
    #   # @return [String] modified buffer
    #   def self.byte_swap_buffer(buffer, num_bytes_per_word)
    #     buffer = buffer.clone
    #     self.byte_swap_buffer!(buffer, num_bytes_per_word)

    # Checks for overflow of an integer data type
    #
    # @param value [Integer] Value to write into the buffer
    # @param min_value [Integer] Minimum allowed value
    # @param max_value [Integer] Maximum allowed value
    # @param hex_max_value [Integer] Maximum allowed value if specified in hex
    # @param bit_size [Integer] Size of the item in bits
    # @param data_type [Symbol] {DATA_TYPES}
    # @param overflow [Symbol] {OVERFLOW_TYPES}
    # @return [Integer] Potentially modified value
    @classmethod
    def check_overflow(cls, value, min_value, max_value, hex_max_value, bit_size, data_type, overflow):
        if overflow == "TRUNCATE":
            # Note this will always convert to unsigned equivalent for signed integers. A little weird but it matches the Ruby implementation.
            value = value % (hex_max_value + 1)
            if value > max_value:
                value = min_value
        else:
            if value > max_value:
                if overflow == "SATURATE":
                    value = max_value
                elif overflow == "ERROR" or value > hex_max_value:
                    raise ValueError(f"value of {value} invalid for {bit_size}-bit {data_type}")
            elif value < min_value:
                if overflow == "SATURATE":
                    value = min_value
                else:
                    raise ValueError(f"value of {value} invalid for {bit_size}-bit {data_type}")
        return value

    # Checks for overflow of an array of integer data types
    #
    # @param values [Array[Integer]] Values to write into the buffer
    # @param min_value [Integer] Minimum allowed value
    # @param max_value [Integer] Maximum allowed value
    # @param hex_max_value [Integer] Maximum allowed value if specified in hex
    # @param bit_size [Integer] Size of the item in bits
    # @param data_type [Symbol] {DATA_TYPES}
    # @param overflow [Symbol] {OVERFLOW_TYPES}
    # @return [Array[Integer]] Potentially modified values
    @classmethod
    def check_overflow_array(cls, values, min_value, max_value, hex_max_value, bit_size, data_type, overflow):
        for index, value in enumerate(values):
            values[index] = cls.check_overflow(
                value,
                min_value,
                max_value,
                hex_max_value,
                bit_size,
                data_type,
                overflow,
            )
        return values

    def enforce_encoding(self):
        return "ASCII-8BIT"

    def enforce_length(self):
        return True

    def enforce_short_buffer_allowed(self):
        return False

    def enforce_derived_write_conversion(self, item):
        return True


# Store the host endianness so that it only has to be determined once
BinaryAccessor.HOST_ENDIANNESS = BinaryAccessor.get_host_endianness()



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.topics.topic import Topic


class SystemEventsTopic(Topic):
    PRIMARY_KEY = "OPENC3__SYSTEM__EVENTS"

    @classmethod
    def update_topic_offsets(cls):
        Topic.update_topic_offsets([cls.PRIMARY_KEY])

    @classmethod
    def write(cls, type, event):
        event["type"] = type
        Topic.write_topic(cls.PRIMARY_KEY, {"event": json.dumps(event)}, "*", 1000)

    @classmethod
    def read(cls):
        for topic, msg_id, msg_hash, redis in Topic.read_topics([cls.PRIMARY_KEY]):
            yield json.loads(msg_hash[b"event"].decode())



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import time
import threading
from openc3.models.metric_model import MetricModel
from openc3.utilities.sleeper import Sleeper
from openc3.top_level import kill_thread


class Metric:
    # The update interval. How often in seconds metrics are updated by this process
    UPDATE_INTERVAL = 5

    # Mutex protecting class variables
    mutex = threading.Lock()

    # Array of instances used to keep track of metrics
    instances = []

    # Thread used to post metrics across all classes
    update_thread = None

    # Sleeper used to delay update thread
    update_sleeper = None

    # Objects with a generate method to be called on each metric cycle (to generate metrics)
    update_generators = []

    def __init__(self, microservice, scope):
        self.scope = scope
        self.microservice = microservice
        self.data = {}
        self.mutex = threading.Lock()

        # Always make sure there is a update thread
        with Metric.mutex:
            Metric.instances.append(self)
            if Metric.update_thread is None:
                Metric.update_thread = threading.Thread(
                    target=self.update_thread_body,
                    daemon=True,
                )
                Metric.update_thread.start()

    def set(self, name, value, type=None, unit=None, help=None, labels=None, time_ms=None):
        with self.mutex:
            if self.data.get(name) is None:
                self.data[name] = {}
            self.data[name]["value"] = value
            if type:
                self.data[name]["type"] = type
            if unit:
                self.data[name]["unit"] = unit
            if help:
                self.data[name]["help"] = help
            if labels:
                self.data[name]["labels"] = labels
            if time_ms:
                self.data[name]["time_ms"] = time_ms

    def set_multiple(self, data):
        with self.mutex:
            self.data = self.data | data

    def update_thread_body(self):
        Metric.update_sleeper = Sleeper()
        while True:
            start_time = time.time()

            with Metric.mutex:
                for generator in Metric.update_generators:
                    generator.generate(Metric.instances[0])

                for instance in Metric.instances:
                    with instance.mutex:
                        json = {}
                        json["name"] = instance.microservice
                        values = instance.data
                        json["values"] = values
                        if len(values) > 0:
                            MetricModel.set(json, scope=instance.scope)

            # Only check whether to update at a set interval
            run_time = time.time() - start_time
            sleep_time = Metric.UPDATE_INTERVAL - run_time
            if sleep_time < 0:
                sleep_time = 0
            if Metric.update_sleeper.sleep(sleep_time):
                break

    def shutdown(self):
        with Metric.mutex:
            try:
                Metric.instances.remove(self)
            except ValueError:
                pass
            if len(Metric.instances) <= 0:
                if Metric.update_sleeper:
                    Metric.update_sleeper.cancel()
                if Metric.update_thread:
                    kill_thread(self, Metric.update_thread)
                Metric.update_thread = None

    def graceful_kill(self):
        pass

    @classmethod
    def add_update_generator(cls, object):
        Metric.update_generators.append(object)



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import re
from openc3.environment import *
from openc3.utilities.authentication import (
    OpenC3Authentication,
    OpenC3KeycloakAuthentication,
)
from openc3.io.json_drb_object import JsonDRbObject
from openc3.io.json_api_object import JsonApiObject

# TODO: ImportError: cannot import name 'DISCONNECT' from partially initialized module 'openc3.script' (most likely due to a circular import) (/usr/lib/python3.11/site-packages/openc3/script/__init__.py)
# from openc3.script import DISCONNECT
import openc3.script


class ApiServerProxy:
    """Provides a proxy to the JsonDRbObject which communicates with the API server"""

    def generate_url(self):
        """pull openc3-cosmos-cmd-tlm-api url from environment variables"""
        hostname = os.environ.get("OPENC3_API_HOSTNAME")
        if not hostname:
            devel = os.environ.get("OPENC3_DEVEL")
            if devel:
                hostname = "127.0.0.1"
            else:
                hostname = "openc3-cosmos-cmd-tlm-api"

        return f"{OPENC3_API_SCHEMA}://{hostname}:{OPENC3_API_PORT}"

    def generate_timeout(self):
        """pull openc3-cosmos-cmd-tlm-api timeout from environment variables"""
        return float(OPENC3_API_TIMEOUT)

    def generate_auth(self):
        """Generate auth object for use with the JsonDRbObject"""
        if OPENC3_API_TOKEN is None and OPENC3_API_USER is None:
            if OPENC3_API_PASSWORD:
                return OpenC3Authentication()
            else:
                return None
        else:
            return OpenC3KeycloakAuthentication(OPENC3_KEYCLOAK_URL)

    def __init__(self):
        """Create a JsonDRbObject connection to the API server"""
        self.json_drb = JsonDRbObject(
            url=self.generate_url(),
            timeout=self.generate_timeout(),
            authentication=self.generate_auth(),
        )

    def __getattr__(self, func):
        """Proxy methods to the API server through the JsonDRbObject"""

        def method(*args, **kwargs):
            if "scope" not in kwargs:
                kwargs["scope"] = OPENC3_SCOPE
            match func:
                case "shutdown":
                    return self.json_drb.shutdown()
                case "request":
                    return self.json_drb.request(*args, **kwargs)
                case _:
                    disconnect = kwargs.pop("disconnect", None)
                    if openc3.script.DISCONNECT:
                        if disconnect:
                            return disconnect
                        else:
                            # The only commands allowed through in disconnect mode are read-only
                            # Thus we allow the get, list, tlm and limits_enabled and subscribe methods
                            if re.compile(
                                r"\w*_get$|^get_\w*|\w*_list$|^list_\w*|^tlm|^limits_enabled$|^subscribe$"
                            ).match(func):
                                return getattr(self.json_drb, func)(*args, **kwargs)
                            else:
                                return None
                    else:
                        return getattr(self.json_drb, func)(*args, **kwargs)

        return method


class ScriptServerProxy:
    """Provides a proxy to the Script Runner API which communicates with the API server"""

    def generate_url(self):
        """pull openc3-cosmos-script-runner-api url from environment variables"""
        hostname = os.environ.get("OPENC3_SCRIPT_API_HOSTNAME")
        if not hostname:
            devel = os.environ.get("OPENC3_DEVEL")
            if devel:
                hostname = "127.0.0.1"
            else:
                hostname = "openc3-cosmos-script-runner-api"
        return f"{OPENC3_SCRIPT_API_SCHEMA}://{hostname}:{OPENC3_SCRIPT_API_PORT}"

    def generate_timeout(self):
        """pull openc3-cosmos-script-runner-api timeout from environment variables"""
        return float(OPENC3_SCRIPT_API_TIMEOUT)

    def __init__(self):
        """Create a JsonApiObject connection to the API server"""
        self.json_api = JsonApiObject(
            url=self.generate_url(),
            timeout=self.generate_timeout(),
            # JsonApiObject calls generate_auth
        )

    def shutdown(self):
        self.json_api.shutdown()

    def request(self, *method_params, **kw_params):
        if "scope" not in kw_params:
            kw_params["scope"] = OPENC3_SCOPE

        # If 'disconnect' is there delete it and return the value
        disconnect = kw_params.pop("disconnect", None)
        if openc3.script.DISCONNECT and disconnect:
            # If they overrode the return value using the disconnect keyword then return that
            return disconnect
        else:
            return self.json_api.request(*method_params, **kw_params)



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.system.system import System


# Base class for all virtual OpenC3 targets which must be implemented by
# a subclass. Provides a framework and helper methods to implement a
# virtual target which can cycle telemetry values and emit telemetry
# packets.
class SimulatedTarget:
    def __init__(self, target_name):
        self.tlm_packets = {}

        # Generate copy of telemetry packets for this target
        for name, packet in System.telemetry.packets(target_name).items():
            self.tlm_packets[name] = packet.clone()

        # Set defaults, template, and id values
        for name, packet in self.tlm_packets.items():
            packet.restore_defaults()
            setattr(packet, "packet_rate", 0)
            for item in packet.id_items:
                packet.write_item(item, item.id_value)

        self.current_cycle_delta = {}

    def set_rates(self):
        raise RuntimeError("Error: set_rates must be implemented by subclass")

    def write(self, packet):
        raise RuntimeError("Error: write must be implemented by subclass")

    def read(self, count_100hz, time):
        return self.get_pending_packets(count_100hz)

    def tick_period_seconds(self):
        return 0.01  # Override this method to optimize

    def tick_increment(self):
        return 1  # Override this method to optimize

    def set_rate(self, packet_name, rate):
        packet = self.tlm_packets[packet_name.upper()]
        if packet is not None:
            packet.packet_rate = rate

    def get_pending_packets(self, count_100hz):
        pending_packets = []

        # Determine if packets are due to be sent and add to pending
        for name, packet in self.tlm_packets.items():
            if packet.packet_rate > 0:
                if (count_100hz % packet.packet_rate) == 0:
                    pending_packets.append(packet)

        return pending_packets

    def cycle_tlm_item(self, packet, item_name, min, max, first_delta):
        packet_name = packet.packet_name
        if self.current_cycle_delta.get(packet_name) is None:
            self.current_cycle_delta[packet_name] = {}
        if self.current_cycle_delta[packet_name].get(item_name) is None:
            self.current_cycle_delta[packet_name][item_name] = first_delta

        current_delta = self.current_cycle_delta[packet_name][item_name]
        current_value = packet.read(item_name)
        updated_value = current_value + current_delta
        if updated_value < min:
            updated_value = min
            self.current_cycle_delta[packet_name][item_name] = -current_delta
        elif updated_value > max:
            updated_value = max
            self.current_cycle_delta[packet_name][item_name] = -current_delta
        packet.write(item_name, updated_value)
        return updated_value



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import os
import re
from openc3.utilities.store import Store


# Tracks the files which are being stored in buckets for data reduction purposes.
# Files are stored in a Redis set by splitting their filenames and storing in
# a set named SCOPE__TARGET__reducer__TYPE, e.g. DEFAULT__INST__reducer__decom
# Where TYPE can be 'decom', 'minute', or 'hour'. 'day' is not necessary because
# day is the final reduction state. As files are reduced they are removed from
# the set. Thus the sets contain the active set of files to be reduced.
class ReducerModel:
    DECOM_BIN_GZ = re.compile(r"__decom\.bin.gz$")
    REDUCED_MINUTE_BIN_GZ = re.compile(r"__reduced_minute\.bin.gz$")
    REDUCED_HOUR_BIN_GZ = re.compile(r"__reduced_hour\.bin.gz$")

    @classmethod
    def add_file(cls, bucket_key: str):
        # Only reduce tlm files
        bucket_key_split = bucket_key.split("/")
        if bucket_key_split[2] == "tlm":
            # bucket_key is formatted like STARTTIME__ENDTIME__SCOPE__TARGET__PACKET__TYPE.bin
            # e.g. 20211229191610578229500__20211229192610563836500__DEFAULT__INST__HEALTH_STATUS__rt__decom.bin
            _, _, scope, target, _ = os.path.basename(bucket_key).split("__")
            if cls.DECOM_BIN_GZ.match(bucket_key):
                return Store.sadd(f"{scope}__{target}__reducer__decom", bucket_key)
            elif cls.REDUCED_MINUTE_BIN_GZ.match(bucket_key):
                return Store.sadd(f"{scope}__{target}__reducer__minute", bucket_key)
            elif cls.REDUCED_HOUR_BIN_GZ.match(bucket_key):
                return Store.sadd(f"{scope}__{target}__reducer__hour", bucket_key)
            # No else clause because add_file is called with raw files which are ignored

    @classmethod
    def rm_file(cls, bucket_key: str):
        _, _, scope, target, _ = bucket_key.split("__")
        if cls.DECOM_BIN_GZ.match(bucket_key):
            return Store.srem(f"{scope}__{target}__reducer__decom", bucket_key)
        elif cls.REDUCED_MINUTE_BIN_GZ.match(bucket_key):
            return Store.srem(f"{scope}__{target}__reducer__minute", bucket_key)
        elif cls.REDUCED_HOUR_BIN_GZ.match(bucket_key):
            return Store.srem(f"{scope}__{target}__reducer__hour", bucket_key)
        else:
            # We should only remove files that were previously in the set
            # Thus if we don't match the bucket_key it is an error
            raise RuntimeError(f"Unknown file {bucket_key}")

    @classmethod
    def all_files(cls, type_: str, target: str, scope: str):
        return Store.smembers(f"{scope}__{target}__reducer__{type_.lower()}").sort()



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import json
from openc3.models.model import Model


class ScopeModel(Model):
    PRIMARY_KEY = "openc3_scopes"

    # NOTE: The following three class methods are used by the ModelController
    # and are reimplemented to enable various Model class methods to work
    @classmethod
    def get(cls, name: str, scope: str = None):
        return super().get(cls.PRIMARY_KEY, name=name)

    @classmethod
    def names(cls, scope: str = None):
        return super().names(cls.PRIMARY_KEY)

    @classmethod
    def all(cls, scope: str = None):
        return super().all(cls.PRIMARY_KEY)

    # END NOTE

    @classmethod
    def from_json(cls, json_data: str | dict):
        """
        Return:
            [Model] Model generated from the passed JSON
        """
        if isinstance(json_data, str):
            json_data = json.loads(json_data)
        if json_data is None:
            raise RuntimeError("json data is nil")
        return cls(**json_data)

    @classmethod
    def get_model(cls, name: str):
        """Calls self.get_model and then from_json to turn the Hash configuration into a Ruby Model object.
        Return:
            [Object|nil] Model object or nil if name not found under primary_key
        """
        json_data = cls.get(name)
        return cls.from_json(json_data) if json_data else None

    def __init__(
        self,
        name: str,
        text_log_cycle_time: int = 600,
        text_log_cycle_size: int = 50_000_000,
        text_log_retain_time: int = None,
        tool_log_retain_time: int = None,
        cleanup_poll_time: int = 900,
        command_authority: bool = False,
        critical_commanding: str = "OFF",
        shard: int = 0,
        updated_at: int = None,
    ):
        super().__init__(
            self.PRIMARY_KEY,
            name=name,
            updated_at=updated_at,
            # This sets the @scope variable which is sort of redundant for the ScopeModel
            # (since its the same as @name) but every model has a @scope
            scope=name,
        )
        self.text_log_cycle_time = text_log_cycle_time
        self.text_log_cycle_size = text_log_cycle_size
        self.text_log_retain_time = text_log_retain_time
        self.tool_log_retain_time = tool_log_retain_time
        self.cleanup_poll_time = cleanup_poll_time
        self.command_authority = command_authority
        self.critical_commanding = str(critical_commanding).upper()
        if len(self.critical_commanding) == 0:
            self.critical_commanding = "OFF"
        if self.critical_commanding not in ["OFF", "NORMAL", "ALL"]:
            raise RuntimeError(f"Invalid value for critical_commanding: {self.critical_commanding}")
        self.shard = shard
        if self.shard is None:
            self.shard = 0
        self.children = []

    # self.return [Hash] JSON encoding of this model
    def as_json(self):
        return {
            "name": self.name,
            "updated_at": self.updated_at,
            "text_log_cycle_time": self.text_log_cycle_time,
            "text_log_cycle_size": self.text_log_cycle_size,
            "text_log_retain_time": self.text_log_retain_time,
            "tool_log_retain_time": self.tool_log_retain_time,
            "cleanup_poll_time": self.cleanup_poll_time,
            "command_authority": self.command_authority,
            "critical_commanding": self.critical_commanding,
            "shard": self.shard,
        }



# Copyright 2025 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from openc3.packets.packet_item import PacketItem
from openc3.config.config_parser import ConfigParser
from openc3.utilities.logger import Logger
from openc3.utilities.extract import hex_to_byte_string, convert_to_value


class PacketItemParser:
    COMMAND = "Command"
    TELEMETRY = "Telemetry"

    # Parses a packet item definition and creates a new PacketItem

    # This number is a little arbitrary but there are definitely issues at
    # 1 million and you really shouldn't be doing anything this big anyway
    BIG_ARRAY_SIZE = 100_000

    # self.param parser [ConfigParser] Configuration parser
    # self.param packet [Packet] The packet the item should be added to
    # self.param cmd_or_tlm [String] Whether this is a command or telemetry packet
    # self.param warnings [Array<String>] Array of warning strings from PacketConfig
    @classmethod
    def parse(cls, parser, packet, cmd_or_tlm, warnings):
        parser = PacketItemParser(parser, warnings)
        parser.verify_parameters(cmd_or_tlm)
        return parser.create_packet_item(packet, cmd_or_tlm)

    # self.param parser [ConfigParser] Configuration parser
    # self.param warnings [Array<String>] Array of warning strings from PacketConfig
    def __init__(self, parser, warnings):
        self.parser = parser
        self.warnings = warnings
        self.usage = self._get_usage()

    def verify_parameters(self, cmd_or_tlm):
        if "ITEM" in self.parser.keyword and cmd_or_tlm == PacketItemParser.COMMAND:
            raise self.parser.error("ITEM types are only valid with TELEMETRY", self.usage)
        elif "PARAMETER" in self.parser.keyword and cmd_or_tlm == PacketItemParser.TELEMETRY:
            raise self.parser.error("PARAMETER types are only valid with COMMAND", self.usage)

        # The usage is formatted with brackets <XXX> around each option so
        # count the number of open brackets to determine the number of options
        max_options = self.usage.count("<")
        # The last two options (description and endianness) are optional
        self.parser.verify_num_parameters(max_options - 2, max_options, self.usage)
        self.parser.verify_parameter_naming(1)  # Item name is the 1st parameter

    def create_packet_item(self, packet, cmd_or_tlm):
        try:
            item_name = self.parser.parameters[0].upper()
            if packet.items.get(item_name):
                msg = f"{packet.target_name} {packet.packet_name} {item_name} redefined."
                Logger.warn(msg)
                self.warnings.append(msg)
            item = PacketItem(
                item_name,
                self._get_bit_offset(),
                self._get_bit_size(),
                self._get_data_type(),
                self._get_endianness(packet),
                self._get_array_size(),
                "ERROR",
            )  # overflow
            if cmd_or_tlm == PacketItemParser.COMMAND:
                item.minimum = self._get_minimum()
                item.maximum = self._get_maximum()
                item.default = self._get_default()
            item.id_value = self._get_id_value(item)
            item.description = self._get_description()
            if self._append():
                item = packet.append(item)
            else:
                item = packet.define(item)
            return item
        except Exception as error:
            raise self.parser.error(error, self.usage)

    def _append(self):
        return "APPEND" in self.parser.keyword

    def _get_data_type(self):
        index = 2 if self._append() else 3
        # _get_data_type is called in usage before we verify
        # the parameters so we must check the length here
        if index >= len(self.parser.parameters):
            return None
        return self.parser.parameters[index].upper()

    def _get_bit_offset(self):
        if self._append():
            return 0
        try:
            return int(self.parser.parameters[1])
        except ValueError as error:
            raise self.parser.error(error, self.usage)

    def _get_bit_size(self):
        index = 1 if self._append() else 2
        try:
            return int(self.parser.parameters[index])
        except ValueError as error:
            raise self.parser.error(error, self.usage)

    def _get_array_size(self):
        if "ARRAY" not in self.parser.keyword:
            return None

        try:
            index = 3 if self._append() else 4
            array_bit_size = int(self.parser.parameters[index])
            items = int(array_bit_size / self._get_bit_size())
            if items >= PacketItemParser.BIG_ARRAY_SIZE:
                warning = (
                    "Performance Issue!\n"
                    f"In {self.parser.filename}:{self.parser.line_number} your definition of=\n"
                    f"{self.parser.line}\n"
                    f"creates an array with {items} elements. Consider creating a BLOCK if this is binary data."
                )
                Logger.warn(warning)
                self.warnings.append(warning)
            return array_bit_size
        except ValueError as error:
            raise self.parser.error(error, self.usage)

    def _get_endianness(self, packet):
        params = self.parser.parameters
        max_options = self.usage.count("<")
        if len(params) >= max_options:
            endianness = params[max_options - 1].upper()
            if endianness != "BIG_ENDIAN" and endianness != "LITTLE_ENDIAN":
                raise self.parser.error(
                    f"Invalid endianness {endianness}. Must be BIG_ENDIAN or LITTLE_ENDIAN.",
                    self.usage,
                )
        else:
            endianness = packet.default_endianness
        return endianness

    def _get_minimum(self):
        if "ARRAY" in self.parser.keyword:
            return None

        data_type = self._get_data_type()
        if data_type == "STRING" or data_type == "BLOCK":
            return None

        index = 3 if self._append() else 4
        if self.parser.parameters[index] == "nil":
            return None
        return ConfigParser.handle_defined_constants(
            convert_to_value(self.parser.parameters[index]),
            self._get_data_type(),
            self._get_bit_size(),
        )

    def _get_maximum(self):
        if "ARRAY" in self.parser.keyword:
            return None

        data_type = self._get_data_type()
        if data_type == "STRING" or data_type == "BLOCK":
            return None

        index = 4 if self._append() else 5
        if self.parser.parameters[index] == "nil":
            return None
        return ConfigParser.handle_defined_constants(
            convert_to_value(self.parser.parameters[index]),
            self._get_data_type(),
            self._get_bit_size(),
        )

    def _convert_string_value(self, index):
        # If the default value is 0x<data> (no quotes), it is treated as
        # binary data. Otherwise, the default value is considered to be a string.
        if (
            self.parser.parameters[index].upper().startswith("0X")
            and f'"{self.parser.parameters[index]}"' not in self.parser.line
            and f"'{self.parser.parameters[index]}'" not in self.parser.line
        ):
            return hex_to_byte_string(self.parser.parameters[index])
        else:
            return self.parser.parameters[index]

    def _get_default(self):
        if "ARRAY" in self.parser.keyword:
            return []

        index = 3 if self._append() else 4
        data_type = self._get_data_type()
        if data_type == "STRING" or data_type == "BLOCK":
            return self._convert_string_value(index)
        else:
            if data_type != "DERIVED":
                return ConfigParser.handle_defined_constants(
                    convert_to_value(self.parser.parameters[index + 2]),
                    self._get_data_type(),
                    self._get_bit_size(),
                )
            else:
                return convert_to_value(self.parser.parameters[index + 2])

    def _get_id_value(self, item):
        if "ID_" not in self.parser.keyword:
            return None
        data_type = self._get_data_type()
        if data_type == "DERIVED":
            raise self.parser.error("DERIVED data type not allowed for Identifier")
        # For PARAMETERS the default value is the ID value
        if "PARAMETER" in self.parser.keyword:
            return item.default

        index = 3 if self._append() else 4
        if data_type == "STRING" or data_type == "BLOCK":
            return self._convert_string_value(index)
        else:
            return ConfigParser.handle_defined_constants(
                convert_to_value(self.parser.parameters[index]),
                self._get_data_type(),
                self._get_bit_size(),
            )

    def _get_description(self):
        max_options = self.usage.count("<")
        if len(self.parser.parameters) >= max_options - 1:
            return self.parser.parameters[max_options - 2]
        return None

    # There are many different usages of the ITEM and PARAMETER keywords so
    # parse the keyword and parameters to generate the correct usage information.
    def _get_usage(self):
        keyword = self.parser.keyword or ""
        usage = f"{keyword} <ITEM NAME> "
        if "APPEND" not in keyword:
            usage += "<BIT OFFSET> "
        usage += self._bit_size_usage()
        usage += self._type_usage()
        if "ARRAY" in keyword:
            usage += "<TOTAL ARRAY BIT SIZE> "
        usage += self._id_usage()
        usage += "<DESCRIPTION (Optional)> <ENDIANNESS (Optional)>"
        return usage

    def _bit_size_usage(self):
        keyword = self.parser.keyword or ""
        if "ARRAY" in keyword:
            return "<ARRAY ITEM BIT SIZE> "
        else:
            return "<BIT SIZE> "

    def _type_usage(self):
        keyword = self.parser.keyword or ""
        # Item type usage is simple so just return it
        if "ITEM" in keyword:
            return "<TYPE: INT/UINT/FLOAT/STRING/BLOCK/DERIVED> "

        # Build up the parameter type usage based on the keyword
        usage = "<TYPE: "
        # ARRAY types don't have min or max or default values
        if "ARRAY" in keyword:
            usage += "INT/UINT/FLOAT/STRING/BLOCK> "
        else:
            try:
                data_type = self._get_data_type()
            except TypeError:
                # If the data type could not be determined set something
                data_type = "INT"
            # STRING and BLOCK types do not have min or max values
            if data_type == "STRING" or data_type == "BLOCK":
                usage += "STRING/BLOCK> "
            else:
                usage += "INT/UINT/FLOAT> <MIN VALUE> <MAX VALUE> "
            # ID Values do not have default values
            if "ID" not in keyword:
                usage += "<DEFAULT_VALUE> "
        return usage

    def _id_usage(self):
        keyword = self.parser.keyword or ""
        if "ID" not in keyword:
            return ""

        if "PARAMETER" in keyword:
            return "<DEFAULT AND ID VALUE> "
        else:
            return "<ID VALUE> "



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.


from openc3.conversions.conversion import Conversion
from openc3.accessors.binary_accessor import BinaryAccessor
from openc3.config.config_parser import ConfigParser


# Performs a generic conversion by evaluating Ruby code
class GenericConversion(Conversion):
    # self.param code_to_eval [String] The Ruby code to evaluate which should
    #   return the converted value
    # self.param converted_type [Symbol] The converted data type. Must be one of
    #   {BinaryAccessor='DATA_TYPES'}
    # self.param converted_bit_size [Integer] The size in bits of the converted
    #   value
    # self.param converted_array_size [Integer] The size in bits of the converted array
    #   value (full size of all items if array):
    def __init__(
        self,
        code_to_eval,
        converted_type=None,
        converted_bit_size=None,
        converted_array_size=None,
    ):
        super().__init__()
        self.code_to_eval = code_to_eval
        if ConfigParser.handle_none(converted_type):
            converted_type = converted_type.upper()
            if converted_type not in BinaryAccessor.DATA_TYPES:
                raise RuntimeError(f"Invalid type {converted_type}")
            self.converted_type = converted_type
        if ConfigParser.handle_none(converted_bit_size):
            self.converted_bit_size = int(converted_bit_size)
        if ConfigParser.handle_none(converted_array_size):
            self.converted_array_size = int(converted_array_size)
        self.params = [code_to_eval, converted_type, converted_bit_size, converted_array_size]

        # Setup multiline eval where the last line defines the return value for eval
        lines = code_to_eval.splitlines()
        exec_lines = lines[0:(len(lines) - 1)]
        self.exec_lines = compile("\n".join(exec_lines), "<string>", "exec")
        self.eval_line = compile(lines[-1], "<string>", "eval")

    def call(self, value, packet, buffer):
        myself = packet  # For backwards compatibility
        if myself:  # Remove unused variable warning for myself
            generic_globals = {"value": value, "myself": myself, "packet": packet, "buffer": buffer}
            exec(self.exec_lines, generic_globals)
            return eval(self.eval_line, generic_globals)

    # self.return [String] The conversion class followed by the code to evaluate
    def __str__(self):
        return self.code_to_eval

    # self.param (see Conversion#to_config)
    # self.return [String] Config fragment for this conversion
    def to_config(self, read_or_write):
        config = f"    GENERIC_{read_or_write}_CONVERSION_START"
        if self.converted_type is not None:
            config += f" {self.converted_type}"
        if self.converted_bit_size is not None:
            config += f" {self.converted_bit_size}"
        if self.converted_array_size is not None:
            config += f" {self.converted_array_size}"
        config += "\n"
        config << self.code_to_eval
        config += f"    GENERIC_{read_or_write}_CONVERSION_END\n"
        return config



# Copyright 2024 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

from ast import literal_eval


class Accessor:
    def __init__(self, packet=None):
        self.packet = packet
        self.args = []

    def read_item(self, item, buffer):
        return self.__class__.class_read_item(item, buffer)

    def write_item(self, item, value, buffer):
        return self.__class__.class_write_item(item, value, buffer)

    def read_items(self, items, buffer):
        result = {}
        for item in items:
            result[item.name] = self.read_item(item, buffer)
        return result

    def write_items(self, items, values, buffer):
        for index, item in enumerate(items):
            self.write_item(item, values[index], buffer)
        return values

    def enforce_encoding(self):
        return "ASCII-8BIT"

    def enforce_length(self):
        return True

    def enforce_short_buffer_allowed(self):
        return False

    def enforce_derived_write_conversion(self, item):
        return True

    @classmethod
    def class_read_item(cls, item, buffer):
        raise RuntimeError("Must be defined by subclass")

    @classmethod
    def class_write_item(cls, item, value, buffer):
        raise RuntimeError("Must be defined by subclass")

    @classmethod
    def class_read_items(cls, items, buffer):
        result = {}
        for item in items:
            result[item.name] = cls.class_read_item(item, buffer)
        return result

    @classmethod
    def class_write_items(cls, items, values, buffer):
        for index, item in enumerate(items):
            cls.class_write_item(item, values[index], buffer)
        return values

    @classmethod
    def convert_to_type(cls, value, item):
        if value is None:
            return None
        match item.data_type:
            case "OBJECT" | "ARRAY":
                pass  # No conversion on complex OBJECT types
            case "STRING" | "BLOCK":
                if item.array_size is not None:
                    if isinstance(value, str):
                        # Thought about using json.loads here but it doesn't
                        # support basic examples like "[2.2, '3', 4]"
                        # Seems it's pretty strict about quotes and escaping
                        value = literal_eval(value)
                    value = [str(x) for x in value]
                else:
                    value = str(value)
            case "INT" | "UINT":
                if item.array_size is not None:
                    if isinstance(value, str):
                        value = literal_eval(value)
                    value = [int(float(x)) for x in value]
                else:
                    value = int(float(value))
            case "FLOAT":
                if item.array_size is not None:
                    if isinstance(value, str):
                        value = literal_eval(value)
                    value = [float(x) for x in value]
                else:
                    value = float(value)
            case _:
                raise TypeError(f"data_type {item.data_type} is not recognized")

        return value



# Copyright 2023 OpenC3, Inc.
# All Rights Reserved.
#
# This program is free software; you can modify and/or redistribute it
# under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation; version 3 with
# attribution addendums as found in the LICENSE.txt
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# This file may also be used under the terms of a commercial license
# if purchased from OpenC3, Inc.

import inspect
import importlib
from openc3.environment import *


# Interface class implemented by each cloud provider: AWS, GCS, Azure
class Bucket:
    # Raised when the underlying bucket does not exist
    class NotFound(Exception):
        pass

    @classmethod
    def getClient(cls):
        if not OPENC3_CLOUD:
            raise RuntimeError("OPENC3_CLOUD environment variable is required")
        # Base is AwsBucket which works with MINIO, Enterprise implements additional
        bucket_class = OPENC3_CLOUD.capitalize() + "Bucket"
        my_module = None
        try:
            my_module = importlib.import_module("." + OPENC3_CLOUD.lower() + "_bucket", "openc3.utilities")
        # If the file doesn't exist try the Enterprise module
        except ModuleNotFoundError:
            my_module = importlib.import_module("." + OPENC3_CLOUD.lower() + "_bucket", "openc3enterprise.utilities")
        return getattr(my_module, bucket_class)()

    def create(self, bucket):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def ensure_public(self, bucket):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def exist(self, bucket):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def delete(self, bucket):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def get_object(self, bucket, key, path=None):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def list_objects(self, bucket, prefix=None, max_request=None, max_total=None):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def list_files(self, bucket, path, only_directories=False, metadata=False):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def put_object(self, bucket, key, body, content_type=None, cache_control=None, metadata=None):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def check_object(self, bucket, key, retries=True):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def delete_object(self, bucket, key):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def delete_objects(self, bucket, keys):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )

    def presigned_request(self, bucket, key, method, internal=True):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method '{inspect.currentframe().f_code.co_name}'"
        )
